from git import Repo
import os
from XBStrapDistro import XBStrapDistro, Package, Source
from XBStrapSQLite import XBStrapSQLite
from pprint import pprint
import _thread
import PySimpleGUI as sg

repo_url = "https://github.com/managarm/bootstrap-managarm.git"
repo_dir = "boostrap-managarm"
distro = XBStrapDistro(repo_dir)
sqlite_database = XBStrapSQLite(distro, "packages.db")

# Main window
package_selector_column = [
    [sg.Text("XBStrap Distro Editor")],
    [sg.Listbox(values=[], key="__PACKAGE_LIST__", expand_y=True)]
]

package_info_column = [
    [sg.Text("Select Package", key="__PACKAGE_NAME__")]
]

main_window_layout = [
    [sg.Column(package_selector_column, expand_y=True),
     sg.VSeperator(),
     sg.Column(package_info_column, expand_y=True, expand_x=True)
     ]
]
main_window: sg.Window

loading_screen_layout = [
    [sg.Text("Loading...")],
    [sg.ProgressBar(100, key="__PROGRESS__", expand_x=True)]
]
loading_screen_window: sg.Window


def update_git_repo():
    if os.path.exists(os.path.join(repo_dir, ".git")):
        repo = Repo(repo_dir)
        assert not repo.bare
        # origin = repo.remotes.origin
        # origin.pull()
    else:
        repo = Repo.clone_from(repo_url, repo_dir)
        assert not repo.bare


def open_loading_window():
    global loading_screen_window
    loading_screen_window = sg.Window("Loading", loading_screen_layout)


init_status = 0


def perform_init():
    global init_status
    print("Initializing git repository")
    update_git_repo()
    init_status = 1
    print("Reading global sources")
    distro.import_global_sources("bootstrap.yml")
    init_status = 2
    print("Reading packages")
    distro.import_packages("bootstrap.yml")
    init_status = 3
    sqlite_database.update_database()
    init_status = 4


def main():
    global loading_screen_window
    global init_status
    global main_window

    perform_init()

    # Open main window
    main_window = sg.Window("XBStrap Distro Editor", main_window_layout, resizable=True, location=(100, 100), size=(400, 400), finalize=True)
    # Fill package list
    main_window["__PACKAGE_LIST__"].update(values=distro.get_package_names())
    while True:
        event, values = main_window.read()
        print(event, values)
        if event == sg.WINDOW_CLOSED or event == 'Quit':
            break

    main_window.close()


if __name__ == '__main__':
    main()



import sqlite3
from dataclasses import dataclass
from XBStrapDistro import XBStrapDistro, Package, Source


@dataclass
class XBStrapSQLiteChangeReport:
    packages: [Package]
    sources: [Source]

    package_change: [str]
    package_changed_source: [str]
    new_packages: [str]

    def get_package_change_reason(self, package_name: str):
        ret = ""
        package_changed: bool = package_name in self.package_change
        package_source_changed: bool = package_name in self.package_changed_source
        package_new: bool = package_name in self.new_packages
        if package_new:
            ret = "New package"
        elif package_changed and package_source_changed:
            ret = "Recipe changed, Source changed"
        else:
            if package_changed:
                ret = "Recipe changed"
            elif package_source_changed:
                ret = "Source changed"
        return ret

    def __init__(self):
        self.packages = []
        self.sources = []
        self.package_change = []
        self.package_changed_source = []
        self.new_packages = []


def update_source_dependencies(c: sqlite3.Cursor, source: Source, change_report: XBStrapSQLiteChangeReport):
    data = []
    for dependency in source.sources_required:
        if type(dependency) is dict:
            is_recursive: bool = False
            if "recursive" in dependency:
                is_recursive = dependency["recursive"]
            data.append((source.name, dependency["name"], is_recursive))
        elif type(dependency) is str:
            data.append((source.name, dependency, False))
        else:
            assert False

    dependencies_in_database = c.execute("SELECT source_name, source_depend, is_recursive FROM source_dependencies"
                                         " WHERE source_name = '{}'"
                                         .format(source.name)).fetchall()

    missing_dependencies = [x for x in data if x not in dependencies_in_database]
    dependencies_to_delete = [x for x in dependencies_in_database if x not in data]
    if len(missing_dependencies):
        c.executemany("INSERT INTO source_dependencies(source_name, source_depend, is_recursive) VALUES(?, ?, ?)",
                      missing_dependencies)
    if len(dependencies_to_delete):
        for dependency in dependencies_to_delete:
            c.execute("DELETE FROM source_dependencies WHERE source_name='{}' AND source_depend='{}'"
                      .format(dependency[0], dependency[1]))


def update_package_dependencies(c: sqlite3.Cursor, package: Package, change_report: XBStrapSQLiteChangeReport):
    data = []
    for dependency in package.dependencies:
        data.append((package.name, dependency))

    dependencies_in_database = c.execute("SELECT package_name, package_depend FROM package_dependencies"
                                         " WHERE package_name = '{}'"
                                         .format(package.name)).fetchall()

    missing_dependencies = [x for x in data if x not in dependencies_in_database]
    dependencies_to_delete = [x for x in dependencies_in_database if x not in data]
    if len(missing_dependencies):
        c.executemany("INSERT INTO package_dependencies(package_name, package_depend) VALUES(?, ?)",
                      missing_dependencies)
    if len(dependencies_to_delete):
        for dependency in dependencies_to_delete:
            print(dependency)
            c.execute("DELETE FROM package_dependencies WHERE package_name='{}' AND package_depend='{}'"
                      .format(dependency[0], dependency[1]))


def update_source_in_database(c: sqlite3.Cursor, source: Source, change_report: XBStrapSQLiteChangeReport) -> bool:
    # Check if row exists
    c.execute("SELECT EXISTS (SELECT 1 FROM sources WHERE source_name='{}' LIMIT 1)".format(source.name))
    if not c.fetchone()[0]:
        change_report.sources.append(source)
        c.execute("INSERT INTO sources(source_name, type, version) VALUES('{}', '{}', '{}')"
                  .format(source.name, source.type, source.version))
        return True
    else:
        c.execute(
            "UPDATE sources SET type='{}', version='{}' WHERE source_name='{}' AND NOT (version='{}' AND type='{}')"
            .format(source.type, source.version, source.name, source.version, source.type))
        if c.rowcount == 1:
            change_report.sources.append(source)
            return True
    return False


def update_package_in_database(c: sqlite3.Cursor, package: Package, change_report: XBStrapSQLiteChangeReport):
    # Check if row exists
    c.execute("SELECT EXISTS (SELECT 1 FROM packages WHERE name='{}' LIMIT 1)".format(package.name))
    if not c.fetchone()[0]:
        c.execute("INSERT OR REPLACE INTO packages(name, source_name, revision) VALUES('{}', '{}', {})"
                  .format(package.name, package.source.name, package.revision))
        change_report.packages.append(package)
        change_report.new_packages.append(package)
    else:
        c.execute(
            "UPDATE packages SET source_name='{}', revision='{}' WHERE name='{}' AND NOT (source_name='{}' AND revision='{}')"
            .format(package.source.name, package.revision, package.name, package.source.name, package.revision))
        if c.rowcount == 1:
            change_report.packages.append(package)
            change_report.package_change.append(package.name)
    if "__embedded_source" in package.source.name:
        if update_source_in_database(c, package.source, change_report):
            if package not in change_report.packages:
                change_report.packages.append(package)
            change_report.package_changed_source.append(package.name)


class XBStrapSQLite:
    distro: XBStrapDistro

    def __initialize_database(self):
        c = self.database.cursor()
        c.execute("CREATE TABLE IF NOT EXISTS packages(name CHAR PRIMARY KEY, source_name CHAR, revision INT)")
        c.execute("CREATE TABLE IF NOT EXISTS package_dependencies(package_name CHAR, package_depend CHAR)")
        c.execute(
            "CREATE TABLE IF NOT EXISTS source_dependencies(source_name CHAR, source_depend CHAR, is_recursive BOOL)")
        c.execute("CREATE TABLE IF NOT EXISTS file_lines(package_name CHAR, entry CHAR, file CHAR, line INT)")
        c.execute("CREATE TABLE IF NOT EXISTS sources(source_name CHAR PRIMARY KEY, type CHAR, version CHAR)")

    def __update_global_sources_in_database(self, change_report: XBStrapSQLiteChangeReport):
        c = self.database.cursor()
        for source in self.distro.global_sources:
            update_source_in_database(c, source, change_report)
        self.database.commit()

    def __update_packages_in_database(self, change_report: XBStrapSQLiteChangeReport):
        c = self.database.cursor()
        for package in self.distro.packages:
            update_package_in_database(c, package, change_report)
        self.database.commit()

    def __update_package_dependencies_in_database(self, change_report: XBStrapSQLiteChangeReport):
        c = self.database.cursor()
        for package in self.distro.packages:
            update_package_dependencies(c, package, change_report)
        self.database.commit()

    def __update_source_dependencies_in_database(self, change_report: XBStrapSQLiteChangeReport):
        c = self.database.cursor()
        for source in self.distro.global_sources:
            update_source_dependencies(c, source, change_report)
        for package in self.distro.packages:
            if package.source not in self.distro.global_sources:
                update_source_dependencies(c, package.source, change_report)
        self.database.commit()

    def update_database(self):
        change_report = XBStrapSQLiteChangeReport()
        print("--- Beginning database update ---")
        print("Updating global sources")
        self.__update_global_sources_in_database(change_report)
        print("Updating packages")
        self.__update_packages_in_database(change_report)
        print("Updating package dependencies")
        self.__update_package_dependencies_in_database(change_report)
        print("Updating source dependencies")
        self.__update_source_dependencies_in_database(change_report)
        print("--- Done doing database update ---")
        return change_report

    def __init__(self, _distro: XBStrapDistro, sql_lite_file):
        self.distro = _distro
        self.database = sqlite3.connect(sql_lite_file)
        assert self.database is not None
        self.__initialize_database()



import os
from pprint import pprint
from dataclasses import dataclass

import yaml
from yaml.loader import SafeLoader

@dataclass
class Metadata:
    categories: []
    line: str = ""
    summary: str = ""
    description: str = ""
    website: str = ""
    maintainer: str = ""
    spdx: str = ""

    def __init__(self, metadata_data):
        self.categories = []
        if metadata_data is None:
            return
        self.line = metadata_data["__line__"]
        if "summary" in metadata_data:
            self.summary = metadata_data["summary"]
        if "description" in metadata_data:
            self.description = metadata_data["description"]
        if "website" in metadata_data:
            self.website = metadata_data["website"]
        if "maintainer" in metadata_data:
            self.maintainer = metadata_data["maintainer"]
        if "categories" in metadata_data:
            self.categories = metadata_data["categories"]
        if "spdx" in metadata_data:
            self.spdx = metadata_data["spdx"]


@dataclass
class Source:
    name: str
    file: str
    line: str
    sources_required: []
    tools_required: []
    regenerate: []
    version: str = "unknown"
    subdir: str = ""
    url: str = ""
    checksum: str = ""
    extract_path: str = ""
    type: str = "unknown"
    tag: str = ""
    branch: str = ""
    commit: str = ""

    def __init__(self, source_data, filename, name=None):
        self.sources_required = []
        self.tools_required = []
        self.regenerate = []
        self.line = source_data["__line__"]
        self.file = filename
        if name is not None:
            self.name = name
        else:
            assert "name" in source_data
            self.name = source_data["name"]
        if "subdir" in source_data:
            self.subdir = source_data["subdir"]
        if "url" in source_data:
            if "format" in source_data:
                self.type = source_data["format"] + "_"
            else:
                self.type = ""
            self.type = self.type + "file"
            self.url = source_data["url"]
        elif "git" in source_data:
            self.type = "git_repository"
            self.url = source_data["git"]
        elif "svn" in source_data:
            self.type = "subversion_repository"
            self.url = source_data["svn"]
        elif "hg" in source_data:
            self.type = "hg_repository"
            self.url = source_data["hg"]
        elif self.subdir == "meta-sources":
            self.type = "metasource"
        if "extract_path" in source_data:
            self.extract_path = source_data["extract_path"]
        if "version" in source_data:
            self.version = source_data["version"]
        if "sources_required" in source_data:
            self.sources_required = source_data["sources_required"]
            if self.type == "unknown":
                self.type = "metasource"
        if "tools_required" in source_data:
            self.tools_required = source_data["tools_required"]
        if "regenerate" in source_data:
            self.regenerate = source_data["regenerate"]
        if "tag" in source_data:
            self.tag = source_data["tag"]
        if "branch" in source_data:
            self.branch = source_data["branch"]
        if "commit" in source_data:
            self.commit = source_data["commit"]
        if self.type == "unknown":
            print(" --- ERROR ---\n  unknown source type!")
            pprint(source_data)
            assert False


@dataclass
class Package:
    name: str
    source: Source
    metadata: Metadata
    labels: []
    dependencies: []
    tools_required: []
    configure: []
    build: []
    revision: int = 1
    architecture: str = ""
    line: str = ""
    file: str = ""
    from_source: str = ""
    stability_level: str = "okay"

    def __init__(self, package_data, filename):
        self.file = filename
        self.line = package_data["__line__"]
        self.labels = []
        self.dependencies = []
        self.tools_required = []
        self.configure = []
        self.build = []
        # Basic package metadata
        if "name" in package_data:
            self.name = package_data["name"]
        if "pkgs_required" in package_data:
            self.dependencies = package_data["pkgs_required"]
        else:
            self.dependencies = []
        if "revision" in package_data:
            self.revision = package_data["revision"]
        if "stability_level" in package_data:
            self.stability_level = package_data["stability_level"]
        if "labels" in package_data:
            self.labels = package_data["labels"]
        if "architecture" in package_data:
            self.architecture = package_data["architecture"]

        # Package buiöd
        if "tools_required" in package_data:
            self.tools_required = package_data["tools_required"]
        if "configure" in package_data:
            self.configure = package_data["configure"]
        if "build" in package_data:
            self.build = package_data["build"]

        # More package metadata
        if "metadata" in package_data:
            self.metadata = Metadata(package_data["metadata"])
        else:
            self.metadata = Metadata(None)

        # Package source
        if "source" in package_data:
            self.source = Source(package_data["source"], filename, self.name + "__embedded_source")
        elif "from_source" in package_data:
            self.from_source = package_data["from_source"]
            self.source = None

class XBStrapDistro:
    class SafeLineLoader(SafeLoader):
        def construct_mapping(self, node, deep=False):
            mapping = super(XBStrapDistro.SafeLineLoader, self).construct_mapping(node, deep=deep)
            # Add 1 so line numbering starts at 1
            mapping['__line__'] = node.start_mark.line + 1
            return mapping

    # TODO: better line number storage

    __repo_dir = ""

    __global_sources: list[Source] = []
    __packages: list[Package] = []

    @property
    def global_sources(self):
        return self.__global_sources


    @property
    def packages(self):
        return self.__packages

    def find_source_by_name(self, name):
        for source in self.__global_sources:
            if source.name == name:
                return source
        return None

    def import_global_sources(self, file_path):
        data = None
        with open(os.path.join(self.__repo_dir, file_path), "r") as stream:
            try:
                data = yaml.load(stream, Loader=self.SafeLineLoader)
            except yaml.YAMLError as exc:
                print(exc)
                return



        if "imports" in data:
            for file_import in data["imports"]:
                self.import_global_sources(file_import["file"])

        if "sources" in data:
            for source_data in data["sources"]:
                self.__global_sources.append(Source(source_data, file_path))

    def import_packages(self, file_path):
        data = None
        with open(os.path.join(self.__repo_dir, file_path), "r") as stream:
            try:
                data = yaml.load(stream, Loader=self.SafeLineLoader)
            except yaml.YAMLError as exc:
                print(exc)
                return

        if "imports" in data:
            for file_import in data["imports"]:
                self.import_packages(file_import["file"])

        if "packages" in data:
            for package_data in data["packages"]:
                package = Package(package_data, file_path)
                if package.from_source != "":
                    package.source = self.find_source_by_name(package.from_source)
                assert package.source is not None
                self.__packages.append(package)

    def get_package_names(self):
        ret: list[str] = []
        for package in self.__packages:
            ret.append(package.name)
        return ret

    def __init__(self, folder):
        self.__repo_dir = folder



from setuptools import setup

setup(
    name='XBDistroTools',
    version='0.1.0',
    packages=['XBStrapDistro', 'XBStrapSQLite'],
    url='https://electrode.codes',
    author='electrodeyt',
    author_email='electrodeyt@gmail.com',
    description='A set of tools to interface with xbstrap distributions.'
)
